// kaldilm/python/csrc/kaldilm.h
//
// Copyright (c)  2020  Xiaomi Corporation (authors: Fangjun Kuang)

#ifndef KALDILM_PYTHON_CSRC_KALDILM_H_
#define KALDILM_PYTHON_CSRC_KALDILM_H_

#include "pybind11/pybind11.h"
namespace py = pybind11;

#endif  // KALDILM_PYTHON_CSRC_KALDILM_H_
